
open Pxp_document
open Pxp_ev_parser
open Pxp_tree_parser
open Pxp_types

open Encodings
open Encodings.Utf8.Pcre

type pxp_node =
  ('a Pxp_document.node Pxp_document.extension as 'a) Pxp_document.node
type pxp_document =
  ('a Pxp_document.node Pxp_document.extension as 'a) Pxp_document.document

let xsd_RE = pcre_regexp "^xsd:"

let has_xsd_prefix s = Pcre.pmatch ~rex:xsd_RE (Utf8.get_str s)

let xsd_namespace = Utf8.mk "http://www.w3.org/2001/XMLSchema"
let xsi_namespace = Utf8.mk "http://www.w3.org/2001/XMLSchema-instance"
let xsd_prefix = Utf8.mk "xsd"
let xsi_prefix = Utf8.mk "xsi"
let add_xsd_prefix =
  let prefix = Utf8.concat xsd_prefix (Utf8.mk ":") in
  fun s -> Utf8.concat prefix s

let schema_ns_prefixes =
  [ xsd_prefix, xsd_namespace; xsi_prefix, xsi_namespace ]

let spec = default_namespace_spec
let new_xsd_config () =
  let ns_manager = new Pxp_dtd.namespace_manager in
  List.iter
    (fun (p, ns) -> ns_manager#add_namespace (Utf8.get_str p) (Utf8.get_str ns))
    schema_ns_prefixes;
  { default_namespace_config with
      Pxp_types.enable_namespace_processing = Some ns_manager
  }

let pxp_node_of ?(config = new_xsd_config ()) src =
  parse_wfcontent_entity config src spec
let pxp_document_of ?(config = new_xsd_config ()) src =
  parse_wfdocument_entity config src spec

let pxp_stream_of_file ?(config = new_xsd_config ()) fname =
  let config = { config with drop_ignorable_whitespace = true } in
  let entity_manager =
    create_entity_manager ~is_document:true config (from_file fname)
  in
  let pull_parser =
    create_pull_parser config
      (`Entry_document [ `Extend_dtd_fully; `Parse_xml_decl ])
      entity_manager
  in
  Stream.from (fun _ -> pull_parser ())

(*
class foo_entity_id = object end
let eid = new foo_entity_id
type to_be_visited =
  | Fully of Value.t  (* xml values still to be visited *)
  | Half of Value.t   (* xml values half visited (i.e. E_start_tag generated) *)
  | Other of Value.t  (* other values *)

let pxp_stream_of_value v =
  let stack = ref [Fully v] in
  let f _ = (* lazy visit of a tree of CDuce XML values, stack keeps track of
            what is still to be visited *)
    (match !stack with
    | (Fully ((Value.Xml (Value.Atom a, attrs, _)) as v)) :: tl ->
	let (ns,a) = Atoms.V.value a in
	assert( ns == Ns.empty );
        let tag_ascii = Encodings.Utf8.to_string a in
        let attrs_ascii =
          List.map (fun (k, v) -> (k, Value.get_string_latin1 v))
            (Value.get_fields attrs)
        in
        let event = Some (E_start_tag (tag_ascii, attrs_ascii, eid)) in
        stack := (Half v) :: tl;
        let children = ref [] in  (* TODO inefficient *)
        let push v = children := v :: !children in
        Value.iter_xml
          (fun pcdata -> push (Other (Value.string_utf8 pcdata)))
          (fun v ->
            match v with
            | (Value.Xml (_, _, _)) as v -> push (Fully v)
            | v -> raise (Invalid_argument "Schema_xml.pxp_stream_of_value"))
          v;
        stack := (List.rev !children) @ !stack;
        event
    | (Half (Value.Xml (Value.Atom a, _, _))) :: tl ->
	let (ns,a) = Atoms.V.value a in
	assert( ns == Ns.empty );
        let tag_ascii = Encodings.Utf8.to_string a in
        let event = Some (E_end_tag (tag_ascii, eid)) in
        stack := tl;
        event
    | (Fully (Value.Xml (_, _, _)))::_ | (Half (Value.Xml (_, _, _)))::_ ->
        failwith "Schema_xml.pxp_stream_of_value: non-atom-tag xml value"
    | (Other v) :: tl ->
        let event = Some (E_char_data (Value.get_string_latin1 v)) in
        stack := tl;
        event
    | [] -> None
    | _ -> assert false)
  in
  Stream.from f
*)

open Printf

let string_of_pxp_event = function
  | E_start_doc (version, dtd) -> "E_start_doc"
  | E_end_doc _ -> "E_end_doc"
  | E_start_tag (name, attlist, _, entity_id) -> sprintf "E_start_tag (%s)" name
  | E_end_tag (name, entity_id) -> sprintf "E_end_tag (%s)" name
  | E_char_data data ->
      sprintf "E_char_data (%s)" (Pcre.replace ~pat:"\n" ~templ:"\\n" data)
  | E_pinstr _ -> "E_pinstr"
  | E_comment data -> "E_comment"
  | E_position (entity, line, col) -> "E_position"
  | E_error exn -> sprintf "E_error (%s)" (Pxp_types.string_of_exn exn)
  | E_end_of_stream -> "E_end_of_stream"
  | _ -> assert false

let rec dump_stream s =
  print_endline (string_of_pxp_event (Stream.next s));
  flush stdout;
  dump_stream s

let dump_stream s = try dump_stream s with Stream.Failure -> ()

  (* peek version that assume the stream isn't at the end *)
let peek s =
  match Stream.peek s with
  | Some v -> v
  | None -> raise Stream.Failure

  (* collect all E_char_data events from a PXP stream and return the
  concatenation of their datas *)
let collect_pcdata s =
  let buf = Buffer.create 1 in
  let rec collect () =
    match peek s with
    | E_char_data d ->
        Buffer.add_string buf d; Stream.junk s; collect ()
    | _ -> Buffer.contents buf
  in
  collect ()

module Pxp_helpers =
  struct

    open Pxp_document
    open Pxp_types

    exception PxpHelpers of exn
    let _raise e = raise (PxpHelpers e)
    let space_RE = pcre_regexp " "

    let _tag_name (n: pxp_node) =
      match n#node_type with
      | T_element tag -> Utf8.mk tag
      | _ -> raise Not_found

    let _has_attribute name (n: pxp_node) =
      try
        match n#attribute name with
        | Value _ -> true
        | _ -> false
      with Not_found -> false

    let _attribute name (n: pxp_node) =
      match n#attribute name with
      | Value v -> Utf8.mk v
      | _ -> raise Not_found

    let _has_element e (n: pxp_node) =
      try ignore (find_element e n); true with Not_found -> false

    let _element e (n: pxp_node): pxp_node = find_element e n
    let _elements e (n: pxp_node): pxp_node list = find_all_elements e n

    let _element' names (n: pxp_node): pxp_node =
      let node = ref None in
      (try
        n#iter_nodes (fun n ->
          (match n#node_type with
          | T_element name when List.mem name names ->
              node := Some n;
              raise Exit
          | _ -> ()))
      with Exit -> ());
      match !node with None -> raise Not_found | Some n -> n

    let _elements' names (n: pxp_node): pxp_node list =
      find_all (fun n ->
        match n#node_type with
        | T_element name when List.mem name names -> true
        | _ -> false) n

    let _has_element' names (n: pxp_node) =
      try ignore (_element' names n); true with Not_found -> false

  end

  (** export Ns.t version of defined namespaces *)

let xsd_namespace = Ns.mk xsd_namespace
let xsi_namespace = Ns.mk xsi_namespace

