exception Error of string
open Ident

(* Pattern algebra *)

type descr
type node
module Node: Custom.T with type t = node

val make: fv -> node
val define: node -> descr -> unit

val constr : Types.t -> descr
val cup    : descr -> descr -> descr
val cap    : descr -> descr -> descr

val times  : node -> node -> descr 
val xml    : node -> node -> descr
val record : label -> node -> descr

val capture : id -> descr
val constant: id -> Types.const -> descr

val id: node -> int
val descr: node -> descr
val fv : node -> fv

(* Pretty-printing *)

module Print : sig
  val print: Format.formatter -> descr -> unit
end

(* Pattern matching: static semantics *)

val accept : node -> Types.Node.t
val filter : Types.t -> node -> (id * Types.Node.t) list


(* Pattern matching: compilation *)

module Compile: sig
  type dispatcher

  type actions =
    | AIgnore of result
    | AKind of actions_kind
  and actions_kind = {
    basic: (Types.t * result) list;
    atoms: result Atoms.map;
    chars: result Chars.map;
    prod: result dispatch dispatch;
    xml: result dispatch dispatch;
    record: record option;
  }
  and record =
    | RecLabel of label * result dispatch dispatch
    | RecNolabel of result option * result option
  and 'a dispatch =
    | Dispatch of dispatcher * 'a array
    | TailCall of dispatcher
    | Ignore of 'a
    | Impossible

  and result = int * source array
  and source = 
    | Catch | Const of Types.const 
    | Left of int | Right of int | Recompose of int * int

  val actions: dispatcher -> actions

  val types_of_codes: dispatcher -> Types.t array

  type 'a rhs = Match of (id * int) list * 'a | Fail
      (* ids are listed in the same order as returned by fv_list,
	 not fv ! *)
	 

  val make_branches : Types.t -> (node * 'a) list -> dispatcher * 'a rhs array

(*  val print_dispatcher: Format.formatter -> dispatcher -> unit *)

  val debug_compile : Format.formatter -> Types.Node.t -> node list -> unit
end

module Compile2 : sig
  type dispatcher

  type source = 
    | SrcCapture
    | SrcLeft | SrcRight
    | SrcCst of Types.const
    | SrcPair of source * source
    | SrcFetchLeft of int
    | SrcFetchRight of int
    | SrcLocal of int

  type push = PushConst of Types.const | PushField | PushCapture

  type result = int * source array
  type actions = 
    | AResult of result
    | AKind of actions_kind
  and actions_kind = {
    basic: (Types.t * result) list;
    atoms: result Atoms.map;
    chars: result Chars.map;
    prod: actions_prod;
    xml: actions_prod;
    record: actions_record;
  }
  and actions_record =
    | RecordLabel of label * dispatcher * record_tr array
    | RecordLabelSkip of label * record_tr
    | RecordResult of result
    | RecordMore of result * result (* nomore, more *)
    | RecordImpossible
  and record_tr = (push list * actions_record) Lazy.t
  and actions_prod =
    | LeftRight of result dispatch dispatch
    | RightLeft of result dispatch dispatch
    | Impossible
  and 'a dispatch =
    | Dispatch of dispatcher * 'a array
    | TailCall of dispatcher
    | Ignore of 'a

  val actions: dispatcher -> actions 

  val make_branches : 
    Types.t -> (node * 'a) list -> 
    dispatcher * (int list * 'a) option array

  val print_dispatcher: Format.formatter  -> dispatcher -> unit
end

val approx :
  descr ->
  Types.descr ->
  id list * (id * Types.Const.t) list

val demo: Format.formatter -> descr -> Types.descr -> unit
val demo_compile: Format.formatter -> Types.descr -> descr list -> unit
